
# coding=utf-8
import re
import logging
from datetime import datetime, timedelta
from collections import defaultdict

from metobs import data as d

from aosstower import station
from aosstower.schema import database

LOG = logging.getLogger(__name__)


class LineParseError(Exception):
    """Error parsing line of record data.
    """
    @classmethod
    def raise_wrapped(cls, exception, msg=None):
        import sys
        traceback = sys.exc_info()[2]
        msg = msg or str(exception)
        raise cls(msg), None, traceback

    def add_winds(self):
        east, north, spd = d.wind_vector_components(float(self['wind_speed']),
                                                    float(self['wind_dir']))

        self['winddir_east'] = '%.3d' % east
        self['winddir_north'] = '%.3d' % north
        self['wind_speed'] = '%.3d' % spd

    def add_altimeter(self, elev=station.ELEVATION):
        self['altimeter'] = '%.3d' % d.altimeter(float(self['pressure']), elev)

    def add_dewpoint(self):
        self['dewpoint'] = '%.3d' % d.dewpoint(float(self['air_temp']),
                                               float(self['rh']))


def _make_record(data):
    for key in data:
        if key == 'stamp':
            continue
        if key in database:
            data[key] = database[key].type(data[key])
    return defaultdict(lambda: float('nan'), data)


class ParserV0(object):

    # maps v0 names to names in schema db
    names = {'ACCURAIN': 'accum_precip',
             'TEMP107_1': 'box_air_temp',
             'TEMP107_2': 'air_temp_2',
             'TEMP107_3': 'air_temp_3',
             'TEMP107_4': 'air_temp_4',
             'TEMP107_5': 'air_temp_5',
             'LI200X': 'solar_flux',
             'RH41372': 'rh',
             'TEMP41372': 'air_temp',
             'CS105': 'box_pressure',
             'PAROSCI': 'pressure',
             'WSPD05305': 'wind_speed',
             'WDIR05305': 'wind_dir',
             'CS10162': 'box_rh',
             'RAIN380M': 'precip'}

    @staticmethod
    def maybe_mine(line):
        return line.startswith('TIME')

    def parse(self, line):
        parts = line.split()
        if len(parts) != 32:
            raise LineParseError("Expected 32 components", line)
        raw_data = {}
        for k1, v1 in {k: v for k, v in zip(parts[0::2], parts[1::2])}.items():
            if k1 == 'TIME':
                continue
            if k1 in self.names:
                raw_data[self.names[k1]] = v1
            else:
                raise LineParseError("Unexpected var: %s" % k1, line)
        try:
            time_str = parts[1]
            unix_time = int(time_str)
            raw_data['stamp'] = datetime.utcfromtimestamp(unix_time)
        except (ValueError, TypeError):
            raise LineParseError("Could not parse stamp", line)
        return _make_record(raw_data)


class ParserV1(object):

    names = ['station_id', 'year', 'doy', 'hhmm', 'sec', 'box_pressure',
             'paro_air_temp_period', 'paro_pressure_period', 'paro_air_temp',
             'pressure', 'paro_cal_sig', 'box_rh', 'box_air_temp',
             'air_temp_2', 'air_temp_3', 'air_temp_4', 'wind_speed', 'wind_dir',
             'rh_shield_freq', 'rh', 'air_temp_6_3m', 'dewpoint',
             'rtd_shield_freq', 'air_temp', 'solar_flux', 'precip',
             'accum_precip', 'altimeter']

    @staticmethod
    def maybe_mine(line):
        return re.search('^\d,\d{4},\d{1,3}', line) is not None

    def _get_stamp(self, data):
        year = int(data['year'])
        doy = int(data['doy'])
        dt = datetime.strptime('{:d}.{:03d}'.format(int(year), int(doy)), '%Y.%j')
        secs = d.hhmm_to_offset(data['hhmm'])
        secs += float(data['sec'])
        secs -= (secs % 5)
        dt += timedelta(seconds=secs)
        return dt

    def parse(self, line):
        parts = line.split(',')
        if len(parts) != 29:
            raise LineParseError("Expected 28 parts", line)
        raw_data = {k: v for k, v in zip(self.names, parts)}
        try:
            raw_data['stamp'] = self._get_stamp(raw_data)
        except (TypeError, ValueError):
            raise LineParseError("Could not parse timesamp", line)
        return _make_record(raw_data)


def read_records(source, error_handler=lambda *a: None):
    """Returns a generator for reading records from `source`. Records are
    checked line-by-line so record line versions may be mixed.
    """
    if hasattr(source, 'readlines'):
        fptr = source
    else:
        fptr = open(source)

    for idx, line in enumerate(fptr.readlines()):
        for parser in [ParserV1(), ParserV0()]:
            if parser.maybe_mine(line):
                try:
                    yield parser.parse(line)
                    break  # forces 'else' to execute
                except LineParseError as err:
                    error_handler(idx + 1, line, err)
        else:
            error_handler(idx + 1, line, RuntimeError("no parser found", line))
